      SUBROUTINE INTEGSPC(FR,p,a,xmax,eps,integ,ierr)

! #####################################################
! #
! # Integrator from a to inf of f(x) that satisfies
! # 1 - f has one maximum in [a,inf)
! # 2 - f goes to zero as x goes to inf
! # 3 - f has most of its area between a and b
! #
! # Input parameters:
! # FR - symbolic name of the integrand function
! # p  - array of parameters needed by FR
! # a  - lower integration limit
! # xmax - approx. location of the maximum/minimum
! # eps - absolute error tolerance 
! #
! # Output arguments:
! # integ - value of the integral
! # ierr - error flag
! #
! # 03/02, Juan Pedro Mellado, UCSD
! #
! #####################################################

      IMPLICIT NONE

#include "types.h"

      TREAL FR,p(*)
      EXTERNAL FR
      TREAL a, eps, xmax
      TINTEGER ierr, ir

      TREAL x, dx, f, xmaxex, bex, fmax, b
      TREAL x1,x2,x3,f1,f2,f3
      TREAL integ, integtail
      TINTEGER n, evals, i

      ierr = 0

      n = 1000
      b = xmax*C_10_R
      dx = (b-a)/M_REAL(n-1)

! Locate maximum/minimum
      x = a
      fmax = ABS(FR(x,p))
      DO i = 2,n
         x = x+dx
         f = ABS(FR(x,p))
         IF (fmax.GT.f) EXIT
         fmax = f
      ENDDO
      IF (i.GE.n) ierr=1
      IF (ierr.NE.0) GO TO 999
      xmaxex = x-dx

! Locate end of main interval: f(bex)=fmax/10
      fmax=fmax/C_10_R
      x = xmaxex
      DO i = 2,n
         x = x+dx
         f = ABS(FR(x,p))
         IF (fmax.GE.f) EXIT
      ENDDO
      IF (i.GE.n) ierr=2
      IF (ierr.NE.0) GO TO 999
      bex = x-dx

! Integrate the main interval: [a,bex)
      x1 = a
      x2 = (a+bex)*C_05_R
      x3 = bex
      f1 = FR(x1,p)
      f2 = FR(x2,p)
      f3 = FR(x3,p)
      evals = 0
      CALL RECURSIMP(FR,p,x1,x2,x3,f1,f2,f3,eps,integ,evals,ir)
      IF (ir.NE.0) ierr=3
      IF (ierr.NE.0) GO TO 999

! Integrate the tail
      DO i = 1,n
         x1 = x3
         f1 = f3
         x3 = C_2_R*x1
         f3 = FR(x3,p)
         x2 = C_05_R*(x1+x3)
         f2 = FR(x2,p)
         evals = 0
         CALL RECURSIMP(FR,p,x1,x2,x3,f1,f2,f3,eps,integtail,evals,ir)
         IF (ir.NE.0) ierr=4
         IF (ierr.NE.0) GO TO 999
         IF (ABS(integtail).LT.eps*ABS(integ)) EXIT
         integ = integ+integtail
      ENDDO
      IF (i.GE.n) ierr=5
      IF (ierr.NE.0) GO TO 999
      
      
 999  RETURN
      END
