! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates c0 (the model coefficient)
! and creates a 2D table where the first index is the
! integral scale, l, and the second is the turbulent
! reynolds number, ret. Nodal values are stored at
! vl and vret as the natural logarithm of the values.
! This is done so the interpolation in the table is 
! logarithmic.

      SUBROUTINE SGS_SPPR_TABLE(ng, remin, dre, nre, vg, vret, c0)

      IMPLICIT NONE

#include "types.h"
      
      TINTEGER ng, nre
      TREAL remin, dre
      TREAL c0(ng, nre)
      TREAL vg(ng)
      TREAL vret(nre)

      TREAL gamma, rg

      TREAL params(5)
      TREAL eps, cl, ret, cl0, beta
      TREAL A0, A1, A2, c34
      TINTEGER niter, nmax, ire, ig
      
      TREAL A,ABSERR,B,EPSABS,EPSREL
      TINTEGER IER,NEVAL
      TREAL WORK
      TINTEGER IWORK,KEY,LAST,LENW,LIMIT
      DIMENSION IWORK(100),WORK(400)

      TREAL EKSP_CL_ROOT, EKSP_A0, EKSP_A1, EKSP_A2
      EXTERNAL EKSP_CL_ROOT, EKSP_A0, EKSP_A1, EKSP_A2
      
! Spectrum low wave number constant (never modified below)
      params(1) = C_4_R

! Spectrum exponential constant (never modified below)
#ifdef SINGLE_PREC
      beta = 5.2e0
#else
      beta = 5.2d0
#endif

! Integration and Iteration constants
      nmax = 256
      eps = C_1EM6_R
      EPSABS = C_1EM8_R
      EPSREL = C_1EM8_R
      KEY = 6
      LIMIT = 100
      LENW = LIMIT*4
      c34 = C_3_R/C_4_R
! 

! Initialization Parameters
! This is the initial value of the cl spectrum
! constant. Its correct value is obtained from
! a Newton Rapson root finding, from integrtion
! of the spectrum to give exact turbulent kinetic
! energy and dissipation.
      cl0 = C_1_R

! Lower integration limit
      A = C_0_R

! Flow Parameters
      ret = remin
      DO ire=1, nre
!    Set reynolds table to log(ret)
         vret(ire) = LOG(ret)

!    Intermediate parameters

!    beta * mu
         params(2) = beta/(ret**c34)

!    initial value of cl
         params(3) = cl0

!    turbulent reynolds number
         params(4) = ret

!    Upper integration limit fixed to 10/(beta*mu) 
!    well in the dissipation subrange.
         B = C_10_R/params(2)

!    Find cl(Ret) function so k and epsilon are exact.
         CALL NEWTON(EKSP_CL_ROOT, params, cl0, nmax, eps, niter, cl)
         
!    Set next initial cl0 to the found root to accelerate convergence
!    of following iterations.
         cl0 = cl
         params(3) = cl

!    Minimum gamma is eta/2 L = (Re**-3/4)/2
         gamma = C_05_R/(ret**c34)
!    Gamma factor is set so last gamma = 1/2 (Delta_f == L)
         rg = EXP(c34*LOG(ret)/M_REAL(ng-1))
         DO ig=1, ng
!       Set integral scale array to log
            vg(ig) = LOG(gamma)

!       Gamma parameter of model
            params(5) = gamma
            
!       Calculate c0(Ret,gama)
            CALL QUADAG(EKSP_A0,params,A,B,EPSABS,EPSREL,KEY,A0,&
                 ABSERR,NEVAL,IER,LIMIT,LENW,LAST,IWORK,WORK)
            CALL QUADAG(EKSP_A1,params,A,B,EPSABS,EPSREL,KEY,A1,&
                 ABSERR,NEVAL,IER,LIMIT,LENW,LAST,IWORK,WORK)
            CALL QUADAG(EKSP_A2,params,A,B,EPSABS,EPSREL,KEY,A2,&
                 ABSERR,NEVAL,IER,LIMIT,LENW,LAST,IWORK,WORK)
            
!       Calculate c0 from root of quadratic equation
            c0(ig,ire) = (-A1+SQRT(A1**2-C_4_R*A0*A2))*C_05_R/A2
            
!       Increase gamma
            gamma = gamma*rg
         ENDDO
!    Increase Ret
         ret = ret*dre
      ENDDO

      RETURN
      END
      
