! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the Turbulent Kinetic Energy
! Spectrum of Pope (2000).
!
! x         = dimensionalized wave number
! p         = low energy power coefficient
! beta_mu   = beta * mu (Pantano and Sarkar 2001)
! cl        = energy containing section functional constant

      FUNCTION EKSP(x, p, beta_mu, cl)

      IMPLICIT NONE

#include "types.h"

      TREAL x, p, beta_mu, cl
      TREAL c53, c2
      TREAL EKSP

      c53 = C_5_R/C_3_R
      c2 = (p+c53)*C_05_R

!  EKSP = (x**p)*EXP(-beta_mu*x)/((x*x+cl)**c2)

      EKSP = ( c53*x/(x+cl)**C_2_R + &
           C_2_R*beta_mu*x*x/(x+cl)**c53 ) *&
           EXP(-beta_mu*x*x)

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the integrand of the a0 integral
!
! x         = dimensionalized wave number
! params    = spectrum parameters
!   1  :  p
!   2  :  beta * mu
!   3  :  cl
!   4  :  <unused here>
!   5  :  gamma

      FUNCTION EKSP_A0(x, params)

      IMPLICIT NONE

#include "types.h"

      TREAL x, params(*)
      TREAL G, S
      TREAL EKSP, EKSP_A0

      S = params(5)*x
!  IF ( S .GT. C_PI_R ) THEN
!     G = C_0_R
!  ELSE IF ( S .GT. C_1EM2_R ) THEN
!     G = SIN(S)/S
!  ELSE
!     G = C_1_R - S*S/C_6_R
!  ENDIF
      IF ( S .GT. C_1EM2_R ) THEN
         G = SIN(S)/S
      ELSE
         G = C_1_R - S*S/C_6_R
      ENDIF

      EKSP_A0 = -EKSP(x, params(1), params(2), params(3))&
           *(C_1_R-G*G)**2

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the integrand of the a1 integral
!
! x         = dimensionalized wave number
! params    = spectrum parameters
!   1  :  p
!   2  :  beta * mu
!   3  :  cl
!   4  :  <unused here>
!   5  :  gamma

      FUNCTION EKSP_A1(x, params)

      IMPLICIT NONE

#include "types.h"

      TREAL x, params(*)
      TREAL G, S, G2
      TREAL EKSP, EKSP_A1

      S = params(5)*x
!  IF ( S .GT. C_PI_R ) THEN
!     G = C_0_R
!  ELSE IF ( S .GT. C_1EM2_R ) THEN
!     G = SIN(S)/S
!  ELSE
!     G = C_1_R - S*S/C_6_R
!  ENDIF
      IF ( S .GT. C_1EM2_R ) THEN
         G = SIN(S)/S
      ELSE
         G = C_1_R - S*S/C_6_R
      ENDIF

      G2 = G*G
      EKSP_A1 = C_2_R*EKSP(x, params(1), params(2), params(3))*G2*&
           (C_1_R-G)*(C_1_R-G2)

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the integrand of the a2 integral
!
! x         = dimensionalized wave number
! params    = spectrum parameters
!   1  :  p
!   2  :  beta * mu
!   3  :  cl
!   4  :  <unused here>
!   5  :  gamma

      FUNCTION EKSP_A2(x, params)

      IMPLICIT NONE

#include "types.h"

      TREAL x, params(*)
      TREAL G, S, G2
      TREAL EKSP, EKSP_A2

      S = params(5)*x
!  IF ( S .GT. C_PI_R ) THEN
!     G = C_0_R
!  ELSE IF ( S .GT. C_1EM2_R ) THEN
!     G = SIN(S)/S
!  ELSE
!     G = C_1_R - S*S/C_6_R
!  ENDIF
      IF ( S .GT. C_1EM2_R ) THEN
         G = SIN(S)/S
      ELSE
         G = C_1_R - S*S/C_6_R
      ENDIF
      
      G2 = G*G
      EKSP_A2 = EKSP(x, params(1), params(2), params(3))*G2*&
           (C_1_R-G)*(C_1_R-G)*(C_1_R-G2)

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates partial derivative with respect
! to cl of the Turbulent Kinetic Energy Spectrum of Pope (2000) 
!
! x         = dimensionalized wave number
! p         = low energy power coefficient
! beta_mu   = beta * mu (Pantano and Sarkar 2001)
! cl        = energy containing section functional constant

      FUNCTION EKSP_CL(x, p, beta, cl)

      IMPLICIT NONE

#include "types.h"

      TREAL x, p, beta, cl
      TREAL c53, c2
      TREAL EKSP_CL

      c53 = C_5_R/C_3_R
      c2 = (p+c53)*C_05_R
      EKSP_CL = -c2*(x**p)*EXP(-beta*x)/((x*x+cl)**(c2+C_1_R))

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the function for the
! Newton-Rapson iteration on cl to enforce that k and 
! epsilon are exact.

      FUNCTION EKSP_CL_ROOT(cl, params)

      IMPLICIT NONE

#include "types.h"

      TREAL EKSP_CL_ROOT
      TREAL cl
      TREAL params(*)

      TREAL A,ABSERR,B,EPSABS,EPSREL,RESULTF,RESULTFP
      TINTEGER IER,NEVAL
      TREAL WORK
      TINTEGER IWORK,KEY,LAST,LENW,LIMIT
      DIMENSION IWORK(100),WORK(400)
      EXTERNAL FEKSP, FEKSP_CL

! Integration Limits 
      A = C_0_R
      B = C_10_R/params(2)

! Convergence constants
      EPSABS = C_1EM8_R
      EPSREL = C_1EM8_R
      KEY = 6
      LIMIT = 100
      LENW = LIMIT*4
!

! Set the current value of cl
      params(3) = cl

      CALL QUADAG(FEKSP,params,A,B,EPSABS,EPSREL,KEY,RESULTF,&
           ABSERR,NEVAL,IER,LIMIT,LENW,LAST,IWORK,WORK)

      IF ( IER .NE. 0 ) THEN
         WRITE(*,*) 'Failure in first inegral (error=',IER,')'
      ENDIF

      CALL QUADAG(FEKSP_CL,params,A,B,EPSABS,EPSREL,KEY,RESULTFP,&
           ABSERR,NEVAL,IER,LIMIT,LENW,LAST,IWORK,WORK)

      IF ( IER .NE. 0 ) THEN
         WRITE(*,*) 'Failure in second inegral (error=',IER,')'
      ENDIF

      EKSP_CL_ROOT = -RESULTF/RESULTFP

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the k/epsilon condition
! for cl.

      FUNCTION FEKSP(x, params)

      IMPLICIT NONE

#include "types.h"

      TREAL x, params(*)
      TREAL EKSP
      TREAL FEKSP

      FEKSP = EKSP(x, params(1), params(2), params(3))*&
           (C_1_R-C_2_R*x*x/params(4))

      RETURN
      END

! Moment Based Partial Reconstruction Model (MBPR Model)
!
! This function calculates the k/epsilon condition partial
! derivative with respect to cl.

      FUNCTION FEKSP_CL(x, params)

      IMPLICIT NONE

#include "types.h"

      TREAL x, params(*)
      TREAL EKSP_CL
      TREAL FEKSP_CL

      FEKSP_CL = EKSP_CL(x, params(1), params(2), params(3))*&
           (C_1_R-C_2_R*x*x/params(4))

      RETURN
      END
