SUBROUTINE CHEM_PETERS1988(rho, T, GM, z1, zh1, h4)

#include "types.h"
#include "dns_error.h"

  USE DNS_GLOBAL
  USE CHEM_GLOBAL
#ifdef PARALLEL
  USE DNS_MPI
#endif

  IMPLICIT NONE

#ifdef PARALLEL
#include "mpif.h"
#endif

  TREAL rho(imax,jmax,kmax)
  TREAL T(imax,jmax,kmax)
  TREAL GM(imax,jmax,kmax)
  TREAL z1(imax,jmax,kmax,*)
  TREAL zh1(imax,jmax,kmax,*)
  TREAL h4(imax,jmax,kmax)

  TINTEGER ij, k
  TINTEGER is, icp, im, ik
  TREAL cda1, cda2, cs(7)
  TREAL dtmp, lnT
  TREAL omegaT
  TREAL c_ch4, c_o2, c_h2o, c_co2, c_co, c_h2, c_h, c_n2
  TREAL c_m, c_ratio, c_o2_h, c_factor
  TREAL yomega(7), csw(8)
  TREAL omegaI, omegaII, omegaIII, omegaIV
  TREAL kr(REAC_MAX)
  TREAL CDLT, CDLT_EPS, CDLT_MAX, TMAX
  TREAL e1, e2, e3, e4, e5, e6

#define k1    kr(1)
#define k5    kr(2)
#define k10   kr(3)
#define k11   kr(4)
#define KII   kr(5)
#define KIV   kr(6)

  IF ( NMECH .NE. 6 ) THEN
     CALL IO_WRITE_ASCII(efile, 'Error: Wrong number of reactions for'&
          //' Peters1988 mechanism')
     CALL DNS_STOP(DNS_ERROR_NBREAC)
  ENDIF

  cda1 = C_1_R/(gama0*mach*mach)
  cda2 = C_1_R/((gama0-C_1_R)*mach*mach)

! Normalized molecular weight
  DO is=1, NSP
     csw(is) = WGHT(is)/WREF
  ENDDO

  cs(1) = - csw(1)*damkohler(inb_scal)
  cs(2) = - csw(2)*damkohler(inb_scal)
  cs(3) = csw(3)*damkohler(inb_scal)
  cs(4) = csw(4)*damkohler(inb_scal)
  cs(5) = csw(5)*damkohler(inb_scal)
  cs(6) = csw(6)*damkohler(inb_scal)
  cs(7) = csw(7)*C_2_R*damkohler(inb_scal)

! This value is obtained from:
! Max Yh2 in flamelet calculation approx 1e-3
! Min density approx 1e-1
! Molecular weight ratio 1/16
! Take then 1% of 1.6e-3
#ifdef SINGLE_PREC
  CDLT = 1.6e-5
  e1 = 6.5e0
  e2 = 6.5e0
  e3 = 1.5e0
  e4 = 0.75e0
  e5 = 0.4e0
  e6 = 0.4e0
#else
  CDLT = 1.6d-5
  e1 = 6.5d0
  e2 = 6.5d0
  e3 = 1.5d0
  e4 = 0.75d0
  e5 = 0.4d0
  e6 = 0.4d0
#endif
  CDLT_EPS = C_10_R*CDLT
  CDLT_MAX = C_3_R*CDLT_EPS

  TMAX = C_2_R

  TGFM = C_0_R
  IF ( damkohler(inb_scal) .EQ. C_0_R ) RETURN

  DO k=1, kmax
     DO ij=1, imax*jmax

        IF ( T(ij,1,k) .GT. TMAX ) THEN

           lnT = LOG(T(ij,1,k))

           DO ik=1,NMECH
              kr(ik) = mech_pre(ik)*&
                   EXP(-mech_tac(ik)/T(ij,1,k)&
                   +mech_beta(ik)*lnT)
           ENDDO

!       Check all mass fractions are positive
           DO is=1, NSP-1
              z1(ij,1,k,is) = MAX(z1(ij,1,k,is),C_0_R)
           ENDDO

           c_h2  = rho(ij,1,k)*z1(ij,1,k,6)/csw(6) 

           IF ( c_h2 .GT. CDLT ) THEN
              IF ( c_h2 .GT. CDLT_MAX ) THEN
                 c_factor = C_1_R
              ELSE
                 c_factor = TANH((c_h2-CDLT)/CDLT_EPS)
                 c_factor = c_factor*(c_h2/CDLT_MAX)**3
              ENDIF

!       Concentrations of intervening species
              c_ch4 = rho(ij,1,k)*z1(ij,1,k,1)/csw(1)
              c_o2  = rho(ij,1,k)*z1(ij,1,k,2)/csw(2)
              c_h2o = rho(ij,1,k)*z1(ij,1,k,3)/csw(3)
              c_co2 = rho(ij,1,k)*z1(ij,1,k,4)/csw(4)
              c_co  = rho(ij,1,k)*z1(ij,1,k,5)/csw(5)
              c_h   = rho(ij,1,k)*z1(ij,1,k,7)/csw(7) 

              dtmp = C_1_R
              DO is=1,NSP-1
                 dtmp = dtmp - z1(ij,1,k,is)
              ENDDO
              c_n2  = rho(ij,1,k)*dtmp/csw(NSP)

!       Compute Reaction Rates Here
              c_m = e1*c_ch4+e2*c_h2o+e3*c_co2+e4*c_co+e5*c_o2&
                   +e6*c_n2+c_h+c_h2
              c_o2_h = c_o2*c_h
              c_ratio = c_h/c_h2

              omegaI = c_factor*k11*c_ch4*c_h
              omegaII = c_factor*k10*c_ratio*(c_co*c_h2o&
                   -c_co2*c_h2/KII)            
              omegaIII = c_factor*k5*c_o2_h*c_m
              omegaIV = c_factor*k1*(c_o2_h-(c_ratio**3)*(c_h2o**2)&
                   /KIV)

              yomega(1) = cs(1)*omegaI
              yomega(2) = cs(2)*omegaIV
              yomega(3) = cs(3)*(C_2_R*omegaIV-omegaI-omegaII)
              yomega(4) = cs(4)*omegaII
              yomega(5) = cs(5)*(omegaI-omegaII)
              yomega(6) = cs(6)*(C_4_R*omegaI+omegaII+omegaIII&
                   -C_3_R*omegaIV)
              yomega(7) = cs(7)*(omegaIV-omegaI-omegaIII)

!       Reaction rate is zero if zero mass fraction of this species
!       and negative reaction rate. Not possible to consume where 
!       there is no reactant.
              DO is=1, NSP-1
                 IF ( z1(ij,1,k,is) .LE. C_0_R .AND. &
                      yomega(is) .LT. C_0_R ) THEN
                    yomega(is) = C_0_R
                 ENDIF
              ENDDO

              omegaT = C_0_R

              DO is=1, NSP-1
                 IF ( T(ij,1,k) .LT. THERMO_TLIM(3,is) ) THEN
                    im = 2
                 ELSE
                    im = 1
                 ENDIF

!          Enthalpy
                 CHW_HI = C_0_R
                 DO icp=NCP_CHEMKIN, 1, -1
                    CHW_HI = CHW_HI*T(ij,1,k)&
                         + THERMO_AI(icp,im,is)/M_REAL(icp)
                 ENDDO
                 CHW_HI = CHW_HI*T(ij,1,k) + THERMO_AI(6,im,is)

                 omegaT = omegaT + (cda1*GM(ij,1,k)*T(ij,1,k)/csw(is)&
                      -cda2*(GM(ij,1,k)-C_1_R)*CHW_HI)*yomega(is)

                 zh1(ij,1,k,is) = zh1(ij,1,k,is) + yomega(is)
              ENDDO
              h4(ij,1,k) = h4(ij,1,k) + omegaT
              TGFM = MAX(TGFM, ABS(omegaT))
           ENDIF
        ENDIF
     ENDDO
  ENDDO

!  PRINT *, TGFM

! Distribute TGFM across processors
#ifdef PARALLEL

  CALL MPI_ALLREDUCE(TGFM, dtmp, 1, MPI_REAL8, MPI_MAX, &
       MPI_COMM_WORLD, ims_err)
  TGFM = dtmp

#endif

  RETURN
END SUBROUTINE CHEM_PETERS1988
