!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2007/01/01 - J.P. Mellado
!#              Created
!#
!########################################################################
!# DESCRIPTION
!#
!# Calculate two first eigenvectors of real symmetric tensor.
!# The third can be computed a posteriori with cross product and it is not 
!# done here to save space.
!# Global orientation is positive Ox for the first one, positive Oy 
!# for the second one.
!#
!########################################################################
!# ARGUMENTS 
!#
!# tensor    In    Array containing the tensor elements: 
!#                 a11, a22, a33, a12, a13, a23
!#           Out   First two normalized eigenvectors (positions 1-3 and 4-6) 
!# lambda    In    Eigenvalues. The third one is not used.
!#
!########################################################################
SUBROUTINE FI_TENSOR_EIGENFRAME(imax, jmax, kmax, tensor, lambda)

  IMPLICIT NONE

#include "types.h"
#include "integers.h"

  TINTEGER imax, jmax, kmax
  TREAL tensor(imax*jmax*kmax,6)
  TREAL lambda(imax*jmax*kmax,*)
  
! -----------------------------------------------------------------------
  TINTEGER ij, il
  TREAL e1(2), e2(2), e3(2), d

! #######################################################################
  DO ij = 1,imax*jmax*kmax
! Loop in number of eigenvalues
     DO il = 1,2

! If two first eqns are l.i.
        d = (tensor(ij,1)-lambda(ij,il))*(tensor(ij,2)-lambda(ij,il))-tensor(ij,4)*tensor(ij,4)
        IF ( d .NE. C_0_R ) THEN
           e1(il) =-( tensor(ij,5)*(tensor(ij,2)-lambda(ij,il)) - tensor(ij,6)*tensor(ij,4) )/d
           e2(il) =-( tensor(ij,6)*(tensor(ij,1)-lambda(ij,il)) - tensor(ij,5)*tensor(ij,4) )/d
           e3(il) = C_1_R
        ELSE
! then two last eqns are l.i.
           d = (tensor(ij,3)-lambda(ij,il))*(tensor(ij,2)-lambda(ij,il))-tensor(ij,6)*tensor(ij,6)
           e1(il) = C_1_R
           e2(il) =-( tensor(ij,4)*(tensor(ij,3)-lambda(ij,il)) - tensor(ij,6)*tensor(ij,5) )/d
           e3(il) =-( tensor(ij,5)*(tensor(ij,2)-lambda(ij,il)) - tensor(ij,6)*tensor(ij,4) )/d
        ENDIF

     ENDDO

! normalize, orient and store in array tensor
     d = sign(C_1_R,e1(1))/SQRT(e1(1)*e1(1)+e2(1)*e2(1)+e3(1)*e3(1)) ! orientation positive Ox
     tensor(ij,3) = e3(1)*d
     tensor(ij,2) = e2(1)*d
     tensor(ij,1) = e1(1)*d
     d = sign(C_1_R,e2(2))/SQRT(e1(2)*e1(2)+e2(2)*e2(2)+e3(2)*e3(2)) ! orientation positive Oy
     tensor(ij,6) = e3(2)*d
     tensor(ij,5) = e2(2)*d
     tensor(ij,4) = e1(2)*d

  ENDDO
     
  RETURN
END SUBROUTINE FI_TENSOR_EIGENFRAME
