!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2007/01/01 - J.P. Mellado
!#              Created
!#
!########################################################################
!# DESCRIPTION
!#
!# Calculate eigenvalues of real symmetric tensor.
!# It assumes that no eigenvalue is zero, i.e. no zero determinant
!#
!########################################################################
!# ARGUMENTS 
!#
!# tensor    In    Array containing the tensor elements: 
!#                 a11, a22, a33, a12, a13, a23
!# result    Out   Array containing eigenvalues: lambda1>lambda2>lambda3
!#
!########################################################################
SUBROUTINE FI_TENSOR_EIGENVALUES(imax, jmax, kmax, tensor, result)

  IMPLICIT NONE

#include "types.h"
#include "integers.h"

  TINTEGER imax, jmax, kmax
  TREAL tensor(imax*jmax*kmax,6)
  TREAL result(imax*jmax*kmax,3)
  
! -----------------------------------------------------------------------
  TINTEGER ij
  TREAL d, e, lambda1, lambda2, lambda3, r13
  TCOMPLEX root1, root2, root3, unit

! #######################################################################
  r13  = C_1_R/C_3_R
  unit = (-0.5d0,0.866025403784439d0)

! #######################################################################
! Define coefficients of characteristic polynomial
! lambda^3 + a lambda^2 + b lambda + c = 0
! result1 contains a, result2 b and result3 c
! #######################################################################
  DO ij = 1,imax*jmax*kmax
     result(ij,1) =-(tensor(ij,1)+tensor(ij,2)+tensor(ij,3))
     result(ij,2) = &
            tensor(ij,1)*tensor(ij,2) + tensor(ij,1)*tensor(ij,3) + tensor(ij,2)*tensor(ij,3) &
          - tensor(ij,4)**2 - tensor(ij,5)**2 - tensor(ij,6)**2
     result(ij,3) = &
          - tensor(ij,1)*(tensor(ij,2)*tensor(ij,3)-tensor(ij,6)*tensor(ij,6)) &
          + tensor(ij,4)*(tensor(ij,4)*tensor(ij,3)-tensor(ij,6)*tensor(ij,5)) &
          + tensor(ij,5)*(tensor(ij,4)*tensor(ij,6)-tensor(ij,2)*tensor(ij,5))
  ENDDO

! #######################################################################
! Solve cubic
! #######################################################################
  DO ij = 1,imax*jmax*kmax
! Depressed eqn (modified) coefficients
     d = result(ij,2) - result(ij,1)*result(ij,1)/C_3_R 
     d = d/C_3_R
     e = result(ij,3) + result(ij,1)*(C_2_R*result(ij,1)*result(ij,1)-C_9_R*result(ij,2))/C_27_R
     e = e/C_2_R

! roots in complex plane
! you could check e**2 + d**3 <= 0, condition for real solns.
     root1 = e**2 + d**3; root1 =-e+sqrt(root1); root1 = root1**r13 ! fundamental root
     root2 = root1*unit
     root3 = root2*unit

     root1 =-d/root1 + root1 - result(ij,1)/C_3_R
     root2 =-d/root2 + root2 - result(ij,1)/C_3_R
     root3 =-d/root3 + root3 - result(ij,1)/C_3_R

! we know they are real
     lambda1 = real(root1)
     lambda2 = real(root2)
     lambda3 = real(root3)

! order from max to min
     d = -result(ij,1) ! trace = lambda1+lambda2+lambda3
     result(ij,1) = max(lambda1,max(lambda2,lambda3))
     result(ij,3) = min(lambda1,min(lambda2,lambda3))
     result(ij,2) = d - result(ij,1) - result(ij,3)

   ENDDO
     
  RETURN
END SUBROUTINE FI_TENSOR_EIGENVALUES
