!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2007/01/01 - J.P. Mellado
!#              Created
!#
!########################################################################
!# DESCRIPTION
!#
!# Strain rate factors for dynamic Smagorinsky model.
!# Moin91 uses the whole rate of strain tensor.
!#
!########################################################################
!# ARGUMENTS 
!#
!# barvs is 1/bar1(bar0(rho))
!#
!# txc1 is bar1(bar0(u) |
!# txc2 is bar1(bar0(v)  > in return
!# txc3 is bar1(bar0(w) |
!#
!# barsij is delta_f^2*bar1(bar0(rho))*SQRT[2*(Sij)_1*bar(Sij)_1]
!# sij    is delta_f^2*bar0(rho)      *SQRT[2(Sij)_0(Sij)_0]
!# 
!# txc size is 5
!#
!########################################################################
SUBROUTINE SGS_DYN_SIJ(n0, n1, x, y, z, dx, dy, dz,&
     rho, u, v, w, sij, barsij, barvs, txc, wrk1d, wrk2d, wrk3d, cf1x, cf1y, cf1z)

#include "types.h"
#include "dns_const.h"
#include "les_const.h"

  USE DNS_GLOBAL
  USE LES_GLOBAL

  IMPLICIT NONE

#include "integers.h"

  TINTEGER n0, n1

  TREAL, DIMENSION(*)                :: x, y, z, dx, dy, dz
  TREAL, DIMENSION(imax,jmax,kmax)   :: u, v, w, rho, sij, barsij, barvs, wrk3d
  TREAL, DIMENSION(imax,jmax,kmax,*) :: txc

  TREAL wrk1d(isize_wrk1d,*)
  TREAL wrk2d(isize_wrk2d,*)

  TREAL, DIMENSION(*) :: cf1x, cf1y, cf1z

! -----------------------------------------------------------------------
  TINTEGER i, j, ij, k

! #######################################################################
! SQRT[2*bar(Sij)_1*bar(Sij)_1]
! #######################################################################
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        txc(ij,1,k,2) = rho(ij,1,k)*u(ij,1,k)
        txc(ij,1,k,3) = rho(ij,1,k)*v(ij,1,k)
        txc(ij,1,k,4) = rho(ij,1,k)*w(ij,1,k)
     ENDDO
  ENDDO
  CALL FILTH(i3, imax, jmax, kmax, kmax_total, &
       iunifx, iunify, iunifz, i1bc, j1bc, k1bc,&
       n1, cf1x, cf1y, cf1z, txc(1,1,1,2), txc(1,1,1,1), wrk3d)
  CALL FILTH(i3, imax, jmax, kmax, kmax_total, &
       iunifx, iunify, iunifz, i1bc, j1bc, k1bc,&
       n1, cf1x, cf1y, cf1z, txc(1,1,1,3), txc(1,1,1,2), wrk3d)
  CALL FILTH(i3, imax, jmax, kmax, kmax_total, &
       iunifx, iunify, iunifz, i1bc, j1bc, k1bc,&
       n1, cf1x, cf1y, cf1z, txc(1,1,1,4), txc(1,1,1,3), wrk3d)
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        txc(ij,1,k,1) = txc(ij,1,k,1)*barvs(ij,1,k)
        txc(ij,1,k,2) = txc(ij,1,k,2)*barvs(ij,1,k)
        txc(ij,1,k,3) = txc(ij,1,k,3)*barvs(ij,1,k)
     ENDDO
  ENDDO

  CALL FI_STRAIN(imode_fdm, imax, jmax, kmax, i1bc, j1bc, k1bc, &
       dx, dy, dz, txc(1,1,1,1), txc(1,1,1,2), txc(1,1,1,3), barsij, &
       txc(1,1,1,4), txc(1,1,1,5), wrk1d, wrk2d, wrk3d)
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        barsij(ij,1,k) = SQRT(C_2_R*barsij(ij,1,k))
     ENDDO
  ENDDO

! #######################################################################
! SQRT[2*bar(Sij)_0*bar(Sij)_0]
! #######################################################################
  CALL FI_STRAIN(imode_fdm, imax, jmax, kmax, i1bc, j1bc, k1bc, &
       dx, dy, dz, u, v, w, sij, &
       txc(1,1,1,4), txc(1,1,1,5), wrk1d, wrk2d, wrk3d)
  DO k = 1,kmax
     DO ij = 1,imax*jmax
        sij(ij,1,k) = SQRT(C_2_R*sij(ij,1,k))
     ENDDO
  ENDDO

! #######################################################################
! Filter size factors
! #######################################################################
  IF ( imode_sim .EQ. DNS_MODE_TEMPORAL ) THEN
     CALL LES_DELTAF3D(imode_sim, i1bc, j1bc, k1bc, imax, jmax,&
          kmax_total, isize_wrk1d, n0, n0, n0, scalex, scaley, &
          scalez, x, y, z, wrk1d(1,1), wrk1d(1,4))
     DO k = 1,kmax
        DO j = 1,jmax
           DO i = 1,imax
              barsij(i,j,k) = barsij(i,j,k)/barvs(i,j,k)*wrk1d(j,1)*wrk1d(j,1)
              sij(i,j,k)    = sij(i,j,k)   *rho(i,j,k)  *wrk1d(j,1)*wrk1d(j,1)
           ENDDO
        ENDDO
     ENDDO

  ELSE IF ( imode_sim .EQ. DNS_MODE_SPATIAL ) THEN
     CALL LES_DELTAF3D(imode_sim, i1bc, j1bc, k1bc, imax, jmax,&
          kmax_total, isize_wrk1d, n0, n0, n0, scalex, scaley, &
          scalez, x, y, z, wrk2d, wrk1d)
     DO k=1,kmax
        DO ij=1,imax*jmax
           barsij(ij,1,k) = barsij(ij,1,k)/barvs(ij,1,k)*wrk2d(ij,1)*wrk2d(ij,1)
           sij(ij,1,k)    = sij(ij,1,k)   *rho(ij,1,k)  *wrk2d(ij,1)*wrk2d(ij,1)
        ENDDO
     ENDDO

  ENDIF

#ifdef _DEBUG
!  CALL IO_WRITE_ASCII_FIELD('Sij.dat',    imax, jmax, kmax, sij   )
!  CALL IO_WRITE_ASCII_FIELD('barSij.dat', imax, jmax, kmax, barsij)
#endif

  RETURN
END SUBROUTINE SGS_DYN_SIJ
