!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 1999/01/01 - C. Pantano
!#              Created
!# 2008/04/17 - J.P. Mellado
!#              Introducing FFTW
!#
!########################################################################
!# DESCRIPTION
!#
!# Normalized s.t. the Parsevals' identity is
!#    1/(nx/2) sum_1^(n/2)E_i \simeq u_rms
!#
!# where the \simeq is due to the 2\delta mode, which is dropped
!#
!########################################################################
!# ARGUMENTS 
!#
!# spectrum  Out  Power spectral density
!# fft_data  In   Array with FFT information, size 2*nx+15 in old fft
!# fft_plan  In   Pointer to FFTW information
!#
!########################################################################
#include "types.h"

SUBROUTINE KX1D_EXECUTE(nx,ny,nz, j, u, fft_plan, fft_data, spectrum)
  
  IMPLICIT NONE

#ifdef PARALLEL
#include "mpif.h"
#endif

  TINTEGER nx,ny,nz
  TREAL u(nx,ny,nz) 
#ifdef USE_FFTW
  TREAL fft_data(nx+2,*)
#else
  TREAL fft_data(2*nx+15,*)
#endif
  TREAL spectrum(nx/2)
  INTEGER(8) fft_plan

! -----------------------------------------------------------------------
  TINTEGER i, j, k

#ifdef PARALLEL
  INTEGER ims_err, itmp
  TINTEGER nz_total
#endif

! #######################################################################
  spectrum(:) = C_0_R

! #######################################################################
! Loop over spanwise direction
! #######################################################################
  DO k = 1,nz
     fft_data(1:nx,1) = u(1:nx,j,k)

#ifdef USE_FFTW
! Perform 1D fft
     CALL dfftw_execute_dft_r2c(fft_plan, fft_data(1,1), fft_data(1,2))
! power spectral density
     spectrum(1) = spectrum(1) + (fft_data(1,2)**2)
     DO i = 2, nx/2
        spectrum(i) = spectrum(i) + fft_data(2*i-1,2)**2 + fft_data(2*i,2)**2
     ENDDO

#else
! Perform 1D fft; fft_data(1,1) is over-written, fft_data(1,2) aux array of size 2*nx+15
     CALL RFFTF(nx, fft_data(1,1), fft_data(1,2))
! power spectral density; the 0-mode contains only real part
     spectrum(1) = spectrum(1) + (fft_data(1,1)**2)
     DO i = 2, nx/2
        spectrum(i) = spectrum(i) + fft_data(2*i-2,1)**2 + fft_data(2*i-1,1)**2
     ENDDO
#endif

  ENDDO

! -----------------------------------------------------------------------
! Final average operation
! -----------------------------------------------------------------------
#ifdef PARALLEL
  CALL MPI_ALLREDUCE(nz, nz_total, 1, MPI_INTEGER4, MPI_SUM, MPI_COMM_WORLD, ims_err)
  itmp = nx/2
  CALL MPI_REDUCE(spectrum, fft_data, itmp, MPI_REAL8, MPI_SUM, 0, MPI_COMM_WORLD, ims_err)

  spectrum(1:nx/2) = fft_data(1:nx/2,1)/M_REAL(nz_total*nx)

#else
  spectrum(1:nx/2) = spectrum(1:nx/2)/M_REAL(nz*nx)

#endif

  RETURN
END SUBROUTINE KX1D_EXECUTE

