#include "types.h"
#include "dns_const.h"
#include "dns_error.h"
#ifdef PARALLEL
#include "dns_const_mpi.h"
#endif

!########################################################################
!# Tool/Library
!#
!########################################################################
!# HISTORY
!#
!# 2014/11/01 - A. de Lozar
!#              Created
!#
!########################################################################
!# DESCRIPTION
!# The transformed field have the inversion point at a fixed height. 
!# The new statisitcs are easier to compare to the 1D turbulence model
!########################################################################
!# ARGUMENTS 
!#
!########################################################################
PROGRAM FIXED_INVERSION

  USE DNS_CONSTANTS
  USE DNS_GLOBAL
#ifdef PARALLEL
  USE DNS_MPI
#endif

  IMPLICIT NONE

#include "integers.h"
#ifdef PARALLEL
#include "mpif.h"
#endif

! Parameter definitions
  TINTEGER, PARAMETER :: itime_size_max = 128

! Fields
  TREAL, DIMENSION(:,:,:,:), ALLOCATABLE, SAVE, TARGET :: q,     s
 TREAL, DIMENSION(:,:,:), ALLOCATABLE, SAVE, TARGET ::    txc ! Array for buoyancy
   

! Work arrays
  TREAL, DIMENSION(:,:), ALLOCATABLE, SAVE :: wrk1d, wrk2d
  TREAL, DIMENSION(:),   ALLOCATABLE, SAVE :: wrk3d

  TINTEGER, DIMENSION(:,:), ALLOCATABLE, SAVE :: inversion_point

  TINTEGER itime_size, it
  TINTEGER itime_vec(itime_size_max)
  TINTEGER isize_wrk3d
  TREAL inv_point_buo
  CHARACTER*32 inifile, bakfile, flow_file, scal_file
  CHARACTER*512 sRes

  TINTEGER i,j,k !Counters
  TINTEGER ref_inversion_point, shift_inv, loc
  TINTEGER dummy

#ifdef PARALLEL
  INTEGER icount
 ! INTEGER ims_err
#endif



! ###################################################################
  inifile = 'dns.ini'
  bakfile = TRIM(ADJUSTL(inifile))//'.bak'

  CALL DNS_INITIALIZE

  CALL DNS_READ_GLOBAL(inifile)

#ifdef PARALLEL
  CALL DNS_MPI_INITIALIZE
#endif


! -------------------------------------------------------------------
! Allocating memory space
! -------------------------------------------------------------------


  ALLOCATE(inversion_point(imax,kmax))
! ------------------------------------------------------------------
 
 isize_wrk3d = MAX(isize_txc_field,imax*jmax*kmax)

! -------------------------------------------------------------------
! File names
! -------------------------------------------------------------------
#ifdef PARALLEL
  IF ( ims_pro .EQ. 0 ) THEN
#endif
     CALL SCANINICHAR(bakfile, inifile, 'PostProcessing', 'Files', '-1', sRes)
     IF ( sRes .EQ. '-1' ) THEN
        WRITE(*,*) 'Integral Iterations ?'
        READ(*,'(A512)') sRes
     ENDIF
     itime_size = itime_size_max
     CALL LIST_INTEGER(sRes, itime_size, itime_vec)
#ifdef PARALLEL
  ENDIF
  CALL MPI_BCAST(itime_size, 1,      MPI_INTEGER4, 0, MPI_COMM_WORLD, ims_err)
  icount = itime_size
  CALL MPI_BCAST(itime_vec,  icount, MPI_INTEGER4, 0, MPI_COMM_WORLD, ims_err)
#endif
  

  ALLOCATE(q(imax,kmax,jmax,inb_flow))
  ALLOCATE(s(imax,jmax,kmax,inb_scal))
  ALLOCATE(txc(imax,jmax,kmax))

  ALLOCATE(wrk3d(isize_wrk3d))
  ALLOCATE(wrk1d(isize_wrk1d,inb_wrk1d))
  ALLOCATE(wrk2d(isize_wrk2d,inb_wrk2d))


 DO it = 1,itime_size
     itime = itime_vec(it)

! main variables
     WRITE(flow_file,*) itime; flow_file=TRIM(ADJUSTL(tag_flow))//TRIM(ADJUSTL(flow_file))

! scalar variables
     WRITE(scal_file,*) itime; scal_file=TRIM(ADJUSTL(tag_scal))//TRIM(ADJUSTL(scal_file))  

  CALL DNS_READ_FIELDS(scal_file, i1, imax,jmax,kmax, inb_scal,i0, isize_wrk3d, s,wrk3d) !Open the 3D fields  
  CALL DNS_READ_FIELDS(flow_file, i2, imax,jmax,kmax, inb_flow,i0, isize_wrk3d, q,wrk3d) !Open the 3D fields

  wrk1d = C_0_R
  CALL FI_BUOYANCY(ibodyforce, imax,jmax,kmax, body_param, s, txc, wrk1d) !Calculate buoyancy in txc
 

  
  inv_point_buo = txc(imax,jmax,kmax)/C_2_R
 
  DO i = 1,imax
     DO k = 1,kmax
        j = jmax
        DO WHILE (txc(i,j,k) .GT. inv_point_buo)
           j = j-1
        ENDDO
        inversion_point(i,k) = j;
     ENDDO
  ENDDO


  
 dummy = MAXVAL(inversion_point(:,:)) ! Find reference inversion point at the top-most value

#ifdef PARALLEL
    CALL MPI_ALLREDUCE(dummy, ref_inversion_point, 1, MPI_INTEGER4, MPI_MAX, MPI_COMM_WORLD, ims_err)
#else
    ref_inversion_point=dummy
#endif



  DO i =1,imax
     DO k = 1,kmax
        shift_inv = ref_inversion_point - inversion_point(i,k) ! Difference between the local inversion point and the reference
      
        DO j = jmax,(1+shift_inv),-1
           loc = j - shift_inv
           s(i,j,k,:) = s(i,loc,k,:)
           q(i,j,k,:) = q(i,loc,k,:)
        END DO
        DO j =shift_inv,2,-1
            loc = 1
           s(i,j,k,:) = s(i,loc,k,:) ! Fill the array with the bottom boundary
           q(i,j,k,:) = q(i,loc,k,:)
           
        END DO

     ENDDO
  ENDDO

  scal_file=TRIM(ADJUSTL(scal_file))//'.trn'
  CALL DNS_WRITE_FIELDS(scal_file, i1, imax,jmax,kmax, inb_scal, isize_wrk3d, s,wrk3d)

  flow_file=TRIM(ADJUSTL(flow_file))//'.trn'
  CALL DNS_WRITE_FIELDS(flow_file, i2, imax,jmax,kmax, inb_flow, isize_wrk3d, q,wrk3d)
  
ENDDO

CALL DNS_END(0)

STOP

END PROGRAM FIXED_INVERSION
